<?php

namespace Drupal\block_usage_report\Controller;

use Drupal\Core\Block\BlockManagerInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityListBuilderInterface;
use Drupal\Core\Entity\EntityPublishedInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns output for the block usage report.
 */
class BlockUsageReportController extends ControllerBase {

  /**
   * The block storage provider.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $blockStorage;

  /**
   * The block content storage provider.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $blockContentStorage;

  /**
   * The block list builder.
   *
   * @var \Drupal\Core\Entity\EntityListBuilderInterface
   */
  protected EntityListBuilderInterface $blockListBuilder;

  /**
   * The block content list builder.
   *
   * @var \Drupal\Core\Entity\EntityListBuilderInterface
   */
  protected EntityListBuilderInterface $blockContentListBuilder;

  /**
   * The fixed block content storage provider, if the associated module exists.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface|null
   */
  protected ?EntityStorageInterface $fixedBlockContentStorage;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $connection;

  /**
   * The block manager service.
   *
   * @var \Drupal\Core\Block\BlockManagerInterface
   */
  protected BlockManagerInterface $blockManager;

  /**
   * Is the block_field module present?
   *
   * @var bool
   */
  protected bool $hasBlockField;

  /**
   * Constructor for this controller.
   *
   * @param \Drupal\Core\Entity\EntityStorageInterface $block_storage
   *   The block storage provider.
   * @param \Drupal\Core\Entity\EntityStorageInterface $block_content_storage
   *   The block content storage provider.
   * @param \Drupal\Core\Entity\EntityListBuilderInterface $block_list_builder
   *   The block list builder.
   * @param \Drupal\Core\Entity\EntityListBuilderInterface $block_content_list_builder
   *   The block content list builder.
   * @param \Drupal\Core\Entity\EntityStorageInterface|null $fixed_block_content_storage
   *   The fixed block content storage provider, or NULL if the associated
   *   module does not exist.
   * @param bool $has_block_field
   *   TRUE if the block_field module is present and enabled, FALSE otherwise.
   */
  public function __construct(
    EntityStorageInterface $block_storage,
    EntityStorageInterface $block_content_storage,
    EntityListBuilderInterface $block_list_builder,
    EntityListBuilderInterface $block_content_list_builder,
    ?EntityStorageInterface $fixed_block_content_storage,
    BlockManagerInterface $block_manager,
    Connection $connection,
    bool $has_block_field
  ) {
    $this->blockStorage = $block_storage;
    $this->blockContentStorage = $block_content_storage;
    $this->blockListBuilder = $block_list_builder;
    $this->blockContentListBuilder = $block_content_list_builder;
    $this->fixedBlockContentStorage = $fixed_block_content_storage;
    $this->blockManager = $block_manager;
    $this->connection = $connection;
    $this->hasBlockField = $has_block_field;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\Core\Extension\ModuleHandlerInterface $module_handler */
    $module_handler = $container->get('module_handler');
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_mgr */
    $entity_type_mgr = $container->get('entity_type.manager');
    /** @var \Drupal\Core\Database\Connection $connection */
    $connection = $container->get('database');
    /** @var \Drupal\Core\Block\BlockManagerInterface $block_mgr */
    $block_mgr = $container->get('plugin.manager.block');

    $fixed_block_storage = NULL;
    if (
      $module_handler->moduleExists('fixed_block_content')
      && $entity_type_mgr->hasHandler('fixed_block_content', 'storage')
    ) {
      $fixed_block_storage = $entity_type_mgr->getStorage('fixed_block_content');
    }

    $has_block_field = $module_handler->moduleExists('block_field');

    return new static(
      $entity_type_mgr->getStorage('block'),
      $entity_type_mgr->getStorage('block_content'),
      $entity_type_mgr->getListBuilder('block'),
      $entity_type_mgr->getListBuilder('block_content'),
      $fixed_block_storage,
      $block_mgr,
      $connection,
      $has_block_field
    );
  }

  /**
   * Finds a content entity referenced by a UUID, given a collection.
   *
   * @param string $uuid
   *   The UUID to search for.
   * @param \Drupal\block_usage_report\Controller\BlockContentPseudoEntity[] $entities
   *   The collection of entities to search for.
   *
   * @return \Drupal\block_usage_report\Controller\BlockContentPseudoEntity|null
   *   THe matching entity, or NULL if not found.
   */
  protected function getEntityByUuid(string $uuid, array $entities): ?BlockContentPseudoEntity {
    foreach ($entities as $entity) {
      if ($entity->uuid() == $uuid) {
        return $entity;
      }
    }
    return NULL;
  }

  /**
   * Builds the output of the report.
   *
   * @return array
   *   The render-array for the report.
   */
  public function build(): array {
    $bc_blocks = BlockContentPseudoEntity::loadMultiple();
    $bc_uuids_found = [];

    /** @var \Drupal\block\Entity\Block[] $blocks */
    $blocks = $this->blockStorage->loadMultiple();
    $public_theme_name = $this->config('system.theme')->get('default');

    $header = [
      $this->t('Name'),
      $this->t('Provider'),
      $this->t('Type'),
      $this->t('Region'),
      $this->t('Operations'),
    ];

    $enabled_rows = $disabled_rows = [];

    foreach ($blocks as $block) {
      $block_theme = $block->getTheme();
      if ($block_theme != $public_theme_name) {
        continue;
      }
      $settings = $block->get('settings');
      $name = $block->label() . ' (' . $block->id() . ')';
      $provider = $settings['provider'];
      $region = $block->getRegion();
      $status = $block->status();
      $plugin_id = $block->getPluginId();
      if (str_contains($plugin_id, ':')) {
        $type = explode(':', $plugin_id)[1];
      }
      else {
        $type = '';
      }

      if ($provider == 'block_content') {
        $uuid = explode(':', $plugin_id)[1];
        $block_content = $this->getEntityByUuid($uuid, $bc_blocks);
        $type = $block_content ? $block_content->bundle() : $this->t('[broken/missing]');
        $bc_uuids_found[] = $uuid;
      }
      elseif (
        $provider == 'fixed_block_content'
        && !empty($type)
        && !empty($this->fixedBlockContentStorage)
        && $fixed_block = $this->fixedBlockContentStorage->load($type)
      ) {
        // Register the fixed block's associated custom block as in-use.
        /** @var \Drupal\fixed_block_content\Entity\FixedBlockContent $fixed_block */
        /** @var \Drupal\block_content\BlockContentInterface $block_content */
        $block_content = $fixed_block->getBlockContent();
        // Show the user the bundle and ID of the associated custom block.
        $type .= '-' . $block_content->bundle() . '-' . $block_content->id();
        $bc_uuids_found[] = $block_content->uuid();
      }

      $row = [
        $name,
        $provider,
        $type,
        $region,
        [
          'data' => [
            '#type' => 'operations',
            '#links' => $this->blockListBuilder->getOperations($block),
          ],
        ],
      ];

      if ($status) {
        $enabled_rows[] = $row;
      }
      else {
        $disabled_rows[] = $row;
      }

    }

    // Sort first by provider, then by type, then by name.
    $sorter = function ($a, $b) {
      if ($a[1] != $b[1]) {
        return $a[1] <=> $b[1];
      }
      if ($a[2] != $b[2]) {
        return $a[2] <=> $b[2];
      }
      return $a[0] <=> $b[0];
    };

    usort($enabled_rows, $sorter);
    usort($disabled_rows, $sorter);

    $fielded_blocks = $this->getFieldedBlockInfo();
    foreach (array_keys($fielded_blocks) as $plugin_id) {
      if (str_starts_with('block_content:', $plugin_id)) {
        $bc_uuids_found[] = substr($plugin_id, 14);
      }
    }

    $unplaced_blocks = [];
    foreach ($bc_blocks as $bc_block) {
      if (!in_array($bc_block->uuid(), $bc_uuids_found)) {
        $bc_block_entity = $this->blockContentStorage->load($bc_block->id());
        $unplaced_blocks[] = [
          $bc_block->label() . ' (' . $bc_block->id() . ')',
          $bc_block->bundle(),
          [
            'data' => [
              '#type' => 'operations',
              '#links' => $this->blockContentListBuilder->getOperations($bc_block_entity),
            ],
          ],
        ];
        // Allow garbage-collection to do its thing.
        unset($bc_block_entity);
      }
    }
    if (!empty($unplaced_blocks)) {
      // Sort first by type, then by name.
      usort($unplaced_blocks, function ($a, $b) {
        if ($a[1] != $b[1]) {
          return $a[1] <=> $b[1];
        }
        return $a[0] <=> $b[0];
      });
    }

    $out = [];
    if (!empty($enabled_rows)) {
      $out['enabled_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Enabled blocks (@ct)', ['@ct' => count($enabled_rows)]),
        '#description' => $this->t('Blocks enabled in regions or field instances.'),
        '#open' => FALSE,
      ];
      $out['enabled_wrapper']['enabled'] = [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $enabled_rows,
      ];
    }
    if (!empty($disabled_rows)) {
      $out['disabled_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Disabled blocks (@ct)', ['@ct' => count($disabled_rows)]),
        '#open' => FALSE,
        '#description' => $this->t('Blocks placed but not enabled in a region.'),
      ];
      $out['disabled_wrapper']['disabled'] = [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $disabled_rows,
      ];
    }

    if (!empty($fielded_blocks)) {
      $out['fielded_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Blocks embedded in content (@ct)', ['@ct' => count($fielded_blocks)]),
        '#open' => FALSE,
        '#description' => $this->t('Blocks embedded in content entities via the block_field module.')
      ];
      $out['fielded_wrapper']['fielded'] = [
        '#theme' => 'table',
        '#header' => [
          $this->t('Name'),
          $this->t('Type'),
          $this->t('Parent Entity Label'),
          $this->t('Entity Type/Bundle'),
          $this->t('Operations'),
        ],
        '#rows' => $fielded_blocks,
      ];
    }

    if (!empty($unplaced_blocks)) {
      $out['unplaced_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Unplaced custom blocks (@ct)', ['@ct' => count($unplaced_blocks)]),
        '#open' => FALSE,
        '#description' => $this->t('Custom blocks not placed in any region or field instance.'),
      ];
      $out['unplaced_wrapper']['unplaced'] = [
        '#theme' => 'table',
        '#header' => [
          $this->t('Name'),
          $this->t('Type'),
          $this->t('Operations'),
        ],
        '#rows' => $unplaced_blocks,
      ];
    }
    return $out;
  }

  /**
   * Returns information about all blocks embedded in block_field instances.
   *
   * @return array[]
   *   Info formatted as rows for table theming.
   */
  protected function getFieldedBlockInfo(): array {
    // Bail if this module is absent.
    if (!$this->hasBlockField) {
      return [];
    }

    // Store our gathered return values here.
    $block_field_info = [];

    // Find all instances of block_field fields.
    /** @var \Drupal\field\Entity\FieldStorageConfig[] $block_fields */
    $block_fields = $this->entityTypeManager()
      ->getStorage('field_storage_config')
      ->loadByProperties(['type' => 'block_field']);

    foreach ($block_fields as $block_field) {
      $field_name = $block_field->getName();
      $entity_type = $block_field->getTargetEntityTypeId();

      // Look up individual instantiations of this field.
      $db_rows = $this->connection->select($entity_type . '__' . $field_name, 'f')
        ->fields('f', ['entity_id', $field_name . '_plugin_id'])
        ->execute()
        ->fetchAll(\PDO::FETCH_NUM);
      foreach ($db_rows as $db_row) {
        [$entity_id, $plugin_id] = $db_row;
        $parent_entity = $this->getOwningEntity($entity_type, $entity_id);
        if (empty($parent_entity) || array_key_exists($plugin_id, $block_field_info)) {
          // Orphaned or malformed, or already found. Skip this one.
          continue;
        }
        if ($parent_entity instanceof EntityPublishedInterface && !$parent_entity->isPublished()) {
          // Omit fielded blocks whose parent entity is unpublished.
          // @todo Revisit this design choice.
          continue;
        }

        $block_def = $this->blockManager->getDefinition($plugin_id);
        // Operations are shown for the owning content entity, NOT the block
        // itself.
        $ops = $this->entityTypeManager()
          ->getListBuilder($parent_entity->getEntityTypeId())
          ->getOperations($parent_entity);

        $block_field_info[$plugin_id] = [
          $block_def['admin_label'] . ' (' . $plugin_id . ')',
          $block_def['provider'],
          ['data' => $parent_entity->toLink()->toRenderable()],
          $parent_entity->getEntityType()->getLabel() . ': ' . $parent_entity->bundle(),
          [
            'data' => [
              '#type' => 'operations',
              '#links' => $ops,
            ],
          ],
        ];
      }
    }

    return $block_field_info;
  }

  /**
   * Finds and loads an entity, escalating up the ownership tree for paragraphs.
   *
   * @param string $entity_type
   *   The entity-type of the current iteration.
   * @param int $entity_id
   *   The entity ID of the current iteration.
   * @param int $revision_id
   *   The revision_id of the current iterations's entity.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The content entity, or NULL in the case of orphaned paragraphs.
   */
  protected function getOwningEntity(string $entity_type, int $entity_id): ?EntityInterface {
    if ($entity_type == 'paragraph') {
      try {
        $parent_data = $this->connection->select('paragraphs_item_field_data', 'p')
          ->fields('p', ['parent_id', 'parent_type'])
          ->condition('id', $entity_id)
          ->execute()
          ->fetch(\PDO::FETCH_NUM);
      }
      catch (\Exception $e) {
        watchdog_exception('block_usage_report', $e);
        $parent_data = NULL;
      }
      if (empty($parent_data)) {
        // This is an orphaned paragraph; ignore.
        // @todo Log this.
        return NULL;
      }
      [$entity_id, $entity_type] = $parent_data;
      if ($entity_type == 'paragraph') {
        return $this->getOwningEntity($entity_type, $entity_id);
      }
    }
    try {
      return $this->entityTypeManager()
        ->getStorage($entity_type)
        ->load($entity_id);
    }
    catch (\Exception $e) {
      watchdog_exception('block_usage_report', $e);
      return NULL;
    }
  }
}

class BlockContentPseudoEntity {

  /**
   * Class constructor. Should never need to be externally instantiated.
   *
   * Note that all arguments are promoted to class properties.
   *
   * @param string $uuid
   *   The block-content UUID.
   * @param string $type
   *   The bundle of the block-content entity.
   * @param string $info
   *   The block-content label.
   * @param int $id
   *   The ID of the block-content entity.
   */
  protected function __construct(
    protected string $uuid,
    protected string $type,
    protected string $info,
    protected int $id
  ) {
    // Nothing here.
  }

  /**
   * Loads and returns info for all block-content entities.
   *
   * @return static[]
   */
  public static function loadMultiple(): array {
    $connection = \Drupal::database();
    $query = $connection->select('block_content', 'bc');
    $query->innerJoin('block_content_field_data', 'bcfd', 'bc.id = bcfd.id');
    $result = $query->fields('bc', ['uuid'])
      ->fields('bcfd', ['id', 'type', 'info'])
      ->execute();
    $blocks = [];
    foreach ($result as $row) {
      $blocks[] = new static(
        $row->uuid,
        $row->type,
        $row->info,
        $row->id
      );
    }
    return $blocks;
  }

  /**
   * Returns the UUID.
   *
   * @return string
   */
  public function uuid(): string {
    return $this->uuid;
  }

  /**
   * Returns the bundle.
   *
   * @return string
   */
  public function bundle(): string {
    return $this->type;
  }

  /**
   * Returns the label.
   *
   * @return string
   */
  public function label(): string {
    return $this->info;
  }

  /**
   * Returns the numeric ID.
   *
   * @return int
   */
  public function id(): int {
    return $this->id;
  }

}